use std::{
    collections::HashSet,
    fs::File,
    io::Read,
    path::{Path, PathBuf},
};

use anyhow::{anyhow, Result};
use clap::{ArgAction, Args, Parser, Subcommand};
use log::{debug, error, info, LevelFilter};
use sha2::{Digest, Sha512};
use simplelog::{ColorChoice, Config, TermLogger, TerminalMode};

use simuldb::{db::transfer, error::Error, neo4rs::ConfigBuilder, prelude::*};

#[derive(Parser, Debug, Clone)]
#[command(author, version, about, multicall = true)]
enum Cli {
    /// Convenience utility to access database information.
    Sdutil(UtilArgs),
    /// Connect to database
    Sddb(DbArgs),
    /// Transfer data from one backend to another
    Sdtransfer(TransferArgs),
}

impl From<Cli> for Command {
    fn from(val: Cli) -> Self {
        match val {
            Cli::Sdutil(UtilArgs { cmd, .. }) => cmd,
            Cli::Sddb(x) => Command::Database(x),
            Cli::Sdtransfer(x) => Command::Transfer(x),
        }
    }
}

#[derive(Args, Debug, Clone)]
struct UtilArgs {
    #[command(flatten)]
    verbosity: Verbosity,
    #[command(subcommand)]
    cmd: Command,
}

#[derive(Args, Debug, Clone)]
struct Verbosity {
    /// Output verbosity
    ///
    /// Specify multiple times to increase verbosity
    #[arg(long, short, action = ArgAction::Count, global = true)]
    verbose: u8,
    /// Disable all logging output
    #[arg(long, short, global = true, conflicts_with = "verbose")]
    quiet: bool,
}

// const LEVELS: [LevelFilter; 6] = [
//     LevelFilter::Off,
//     LevelFilter::Error,
//     LevelFilter::Warn,
//     LevelFilter::Info,
//     LevelFilter::Debug,
//     LevelFilter::Trace,
// ];
const LEVELS: [LevelFilter; 3] = [LevelFilter::Info, LevelFilter::Debug, LevelFilter::Trace];

impl Verbosity {
    fn level_filter(&self) -> LevelFilter {
        if self.quiet {
            LevelFilter::Off
        } else {
            let verbose = self.verbose.clamp(0, LEVELS.len() as u8 - 1) as usize;
            LEVELS[verbose]
        }
    }
}

#[derive(Subcommand, Debug, Clone)]
enum Command {
    /// Connect to database
    Database(DbArgs),
    /// Transfer data from one backend to another
    Transfer(TransferArgs),
}

impl Command {
    fn verbosity(&self) -> &Verbosity {
        match self {
            Command::Database(DbArgs { verbosity, .. }) => verbosity,
            Command::Transfer(TransferArgs { verbosity, .. }) => verbosity,
        }
    }
}

#[derive(Args, Debug, Clone)]
struct DbArgs {
    #[command(flatten)]
    verbosity: Verbosity,
    /// Database connection string
    ///
    /// Should be one of the following
    /// - json://PATH for a JSON based database
    ///     PATH denotes the path to the folder containing the JSON files.
    /// - neo4j://USER:PASS@URI for a Neo4j based database
    ///     URI denotes the connection URI (e.g. localhost:7687).
    ///     USER and PASS are user and password for the connection. If not given, neo4j is used for either of them.
    ///     This can be achieved with neo4j://URI or neo4j://USER@URI.
    #[arg(verbatim_doc_comment)]
    db: String,
    #[command(subcommand)]
    cmd: DbCommand,
}

#[derive(Subcommand, Debug, Clone)]
enum DbCommand {
    /// List session in the database
    List,
    /// Verify hashes of the data files
    ///
    /// This assumes that the filenames correspond to the id and that SHA-512 hashes are used.
    Verify {
        /// Path to folder with data files
        path: PathBuf,
    },
    /// Prune database
    ///
    /// This removes dangling files and references without a corresponding partner.
    /// This assumes that the filenames correspond to the id and that SHA-512 hashes are used.
    Prune {
        /// Path to folder with data files
        path: PathBuf,
    },
}

#[derive(Args, Debug, Clone)]
struct TransferArgs {
    #[command(flatten)]
    verbosity: Verbosity,
    /// Database to copy from
    ///
    /// Should be one of the following
    /// - json://PATH for a JSON based database
    ///     PATH denotes the path to the folder containing the JSON files.
    /// - neo4j://USER:PASS@URI for a Neo4j based database
    ///     URI denotes the connection URI (e.g. localhost:7687).
    ///     USER and PASS are user and password for the connection. If not given, neo4j is used for either of them.
    ///     This can be achieved with neo4j://URI or neo4j://USER@URI.
    #[arg(verbatim_doc_comment)]
    from: String,
    /// Database to copy to
    ///
    /// See <FROM> for more details
    #[arg(verbatim_doc_comment)]
    to: String,
}

fn list<D>(db: &D) -> Result<()>
where
    D: Database,
    D::Error: 'static + Send + Sync,
{
    let mut sessions = db.get_sessions()?;
    sessions.sort_by_key(|x| x.session().run.date.clone());
    for session in sessions {
        let num = session.get_datasets()?.len();
        let session = session.session();

        println!(
            "{:>10} | {} ({}) [{}]: {}",
            num, session.software.name, session.software.version, session.run.id, session.run.date,
        );
    }
    Ok(())
}

fn verify<D, P: AsRef<Path>>(db: &D, data_folder: P) -> Result<()>
where
    D: Database,
    D::Error: 'static + Send + Sync,
{
    let mut total = 0;
    let mut failures = 0;
    for session in db.get_sessions()? {
        for dataset in session.get_datasets()? {
            let mut path = PathBuf::from(data_folder.as_ref());
            path.push(format!("{}", dataset.id));
            info!("Checking {}", path.display());

            let hash = {
                let mut hasher = Sha512::new();
                let mut buffer = Vec::new();
                File::open(&path)?.read_to_end(&mut buffer)?;
                hasher.update(buffer);
                hasher.finalize().as_slice().to_vec()
            };
            total += 1;
            if dataset.hash != hash {
                error!("Hash mismatch for {}", path.display());
                failures += 1;
            }
        }
    }
    println!("{failures}/{total} checks failed");
    Ok(())
}

fn prune<D, P: AsRef<Path>>(db: &D, data_folder: P) -> Result<()>
where
    D: Database,
    D::Error: 'static + Send + Sync,
{
    let data_folder = data_folder.as_ref();
    info!("Pruning {}", data_folder.display());
    let mut referenced_files = HashSet::new();
    for session in db.get_sessions()? {
        info!("Searching {:?}", session.session());
        for dataset in session.get_datasets()? {
            let file_hash = dataset.id.to_string();
            debug!("Found {file_hash}");
            referenced_files.insert(file_hash);
        }
    }
    for entry in data_folder.read_dir()? {
        let entry = entry?;
        let file_path = entry.path();
        if let Some(file_name) = file_path.file_name().and_then(|x| x.to_str()) {
            if !referenced_files.contains(file_name) {
                info!("Removing {}", file_path.display());
                std::fs::remove_file(&file_path)?;
            }
        }
    }

    Ok(())
}

fn parse_db_string(s: String) -> Result<Box<dyn Database<Error = Error>>> {
    match s.split_once("://") {
        Some(("json", path)) => Ok(Box::new(Json::new(path))),
        Some(("neo4j", path)) => {
            let mut config = ConfigBuilder::new();
            match path.split_once('@') {
                None => config = config.uri(path).user("neo4j").password("neo4j"),
                Some((user_pass, uri)) => {
                    config = config.uri(uri);
                    match user_pass.split_once(':') {
                        None => config = config.user(user_pass).password("neo4j"),
                        Some((user, pass)) => config = config.user(user).password(pass),
                    }
                }
            };
            Ok(Box::new(Neo4j::new(config.build()?)?))
        }
        Some((prefix, path)) => Err(anyhow!("Invalid schema {prefix} with path {path}")),
        None => Err(anyhow!("Invalid connection string")),
    }
}

fn main() -> Result<()> {
    let cli: Command = Cli::parse().into();
    let verbosity = cli.verbosity();
    TermLogger::init(
        verbosity.level_filter(),
        Config::default(),
        TerminalMode::Mixed,
        ColorChoice::Auto,
    )?;

    match cli {
        Command::Database(DbArgs { db, cmd, .. }) => {
            let db = parse_db_string(db)?;
            match cmd {
                DbCommand::List => list(&db)?,
                DbCommand::Verify { path } => verify(&db, path)?,
                DbCommand::Prune { path } => prune(&db, path)?,
            }
        }
        Command::Transfer(TransferArgs { from, to, .. }) => {
            let db1 = parse_db_string(from)?;
            let mut db2 = parse_db_string(to)?;
            transfer(&db1, &mut db2).map_err(|x| anyhow!("{x}"))?;
        }
    };
    Ok(())
}
